 MODULE math_util

 real, parameter, private :: EPSS = 1.0E-05

 private :: equal_vector_r, equal_vector_i
 interface equal_vector
     module procedure equal_vector_r, equal_vector_i
 end interface

 private :: equal_matrix_r, equal_matrix_i
 interface equal_matrix
     module procedure equal_matrix_r, equal_matrix_i
 end interface

 private :: is_integer_noeps, is_integer_eps
 interface is_integer
     module procedure is_integer_noeps, is_integer_eps
 end interface

 contains

   logical function is_nan(x)
#ifdef __INTEL_COMPILER
   USE, intrinsic :: ieee_arithmetic
   real, intent(in) :: x
   is_nan = ieee_is_nan(x) 
#else
   USE strutil
   real, intent(in) :: x
   is_nan = (trim(r_to_s(x)) == 'NaN')
#endif
   end function is_nan

!---------------------------------------------------------------------------------------------------------

   logical function is_inf(x)
#ifdef __INTEL_COMPILER
   USE, intrinsic :: ieee_arithmetic
   real, intent(in) :: x
   is_inf = .not.ieee_is_finite(x)
#else
!!!TODO
#endif
   end function is_inf

!---------------------------------------------------------------------------------------------------------
  
   function derivative(a)  result(der)
!
!  First derivative of vector using 2-point central difference
!
   real, dimension(:), intent(in) :: a
   real, dimension(size(a))       :: der
   integer                        :: n
!
   n = size(a)
   der(1) = a(2) - a(1)
   der(n) = a(n) - a(n-1)
!
   der(2:n-1) = a(3:n) - a(1:n-2) / 2
!
   end function derivative

!---------------------------------------------------------------------------------------------------------

   function straight_line(x1,y1,x2,y2) result(line)
!
!  Calculate straigt line between 2 points: y=line(1)*x + line(2)
!
   real, intent(in)   :: x1,y1,x2,y2
   real, dimension(2) :: line
!
   line(1) = (y2 -y1) / (x2 -x1)
   line(2) = -line(1)*x1 + y1
!
   end function straight_line

!---------------------------------------------------------------------------------------------------------

   function line_line_intersection(l1,l2,ier)  result(xp)
!
!  Intersection of two lines l1(1)x+l1(2), l2(1)x+l2(2)
!
   real, dimension(2), intent(in) :: l1,l2
   integer, intent(out)           :: ier
   real, dimension(2)             :: xp
!
   if (abs(l1(1) - l2(1)) <= epsilon(1.0)) then  ! parallel lines
       ier = 1
       xp(:) = 0
   else
       ier = 0
       xp(1) = (l2(2) - l1(2)) / (l1(1) - l2(1))
       xp(2) = (l1(1)*l2(2) - l2(1)*l1(2)) / (l1(1) - l2(1))
   endif
!
   end function line_line_intersection

!---------------------------------------------------------------------------------------------------------

   real function integrate(x,y,n1,n2)  result(area)
!
!  Integrate array in range n1-n2 with n2 >= n1
!
   real, dimension(:), intent(in) :: x,y
   integer, intent(in)            :: n1,n2
!
   if (n1 == n2) then
       area = y(n1)
   else
       area = 0.5 * sum((x(n1+1:n2) - x(n1:n2-1)) * (y(n1+1:n2) + y(n1:n2-1)))
   endif
!
   end function integrate

!---------------------------------------------------------------------------

   logical function equal_matrix_r(mat1,mat2)
!
!  mat1 is equal to mat2 ?
!
   real, dimension(:,:), intent(in) :: mat1,mat2
   equal_matrix_r = all (abs(mat1 - mat2) < EPSS)
   end function equal_matrix_r

!---------------------------------------------------------------------------

   logical function equal_matrix_i(mat1,mat2)
!
!  mat1 is equal to mat2 ?
!
   integer, dimension(:,:), intent(in) :: mat1,mat2
   equal_matrix_i = all (abs(mat1 - mat2) == 0)
   end function equal_matrix_i

!---------------------------------------------------------------------------

   logical function equal_vector_r(vet1,vet2)
!
!  vet1 is equal to vet2 ?
!
   real, dimension(:), intent(in) :: vet1,vet2
   equal_vector_r = all (abs(vet1 - vet2) < EPSS)
   end function equal_vector_r

!---------------------------------------------------------------------------

   logical function equal_vector_i(vet1,vet2)
   integer, dimension(:), intent(in) :: vet1,vet2
   equal_vector_i = all (abs(vet1 - vet2) == 0)
   end function equal_vector_i

!------------------------------------------------------------------------------------------

   elemental logical function is_integer_noeps(rnum)
   real, intent(in) :: rnum
   is_integer_noeps = ceiling(rnum) == rnum
   end function is_integer_noeps

!------------------------------------------------------------------------------------------

   elemental logical function is_integer_eps(rnum,eps)
   real, intent(in) :: rnum,eps
   is_integer_eps = abs(nint(rnum) - rnum) <= eps
   end function is_integer_eps

!------------------------------------------------------------------------------------------

   real function determinante_22(A)
!
   real, dimension(2,2) :: A
!
   determinante_22 = A(1,1)*A(2,2) - A(1,2)*A(2,1) 
!
   end function determinante_22

!------------------------------------------------------------------------------------------

   real function determinante(mat)
!
   real, dimension(3,3) :: mat
!
   determinante = mat(1,1)*mat(2,2)*mat(3,3) + mat(1,2)*mat(2,3)*mat(3,1) + mat(1,3)*mat(2,1)*mat(3,2) &
                - mat(3,1)*mat(2,2)*mat(1,3) - mat(3,2)*mat(2,3)*mat(1,1) - mat(3,3)*mat(2,1)*mat(1,2)
!
   end function determinante

!------------------------------------------------------------------------------------------

   real function determinante_44(A)
!
   real, dimension(4,4) :: A
!
   determinante_44 =   &
     A(1,1)*(A(2,2)*(A(3,3)*A(4,4)-A(3,4)*A(4,3))+A(2,3)*(A(3,4)*A(4,2)-A(3,2)*A(4,4))+A(2,4)*(A(3,2)*A(4,3)-A(3,3)*A(4,2)))&
   - A(1,2)*(A(2,1)*(A(3,3)*A(4,4)-A(3,4)*A(4,3))+A(2,3)*(A(3,4)*A(4,1)-A(3,1)*A(4,4))+A(2,4)*(A(3,1)*A(4,3)-A(3,3)*A(4,1)))&
   + A(1,3)*(A(2,1)*(A(3,2)*A(4,4)-A(3,4)*A(4,2))+A(2,2)*(A(3,4)*A(4,1)-A(3,1)*A(4,4))+A(2,4)*(A(3,1)*A(4,2)-A(3,2)*A(4,1)))&
   - A(1,4)*(A(2,1)*(A(3,2)*A(4,3)-A(3,3)*A(4,2))+A(2,2)*(A(3,3)*A(4,1)-A(3,1)*A(4,3))+A(2,3)*(A(3,1)*A(4,2)-A(3,2)*A(4,1))) 
!
   end function determinante_44

!------------------------------------------------------------------------------------------

   function matinv2(A,ok_flag) result(B)
   !
   !Performs a direct calculation of the inverse of a 3×3 matrix.
   real, intent(in)               :: A(2,2)   ! Matrix
   logical, intent(out), optional :: ok_flag  ! .TRUE. if the input matrix could be inverted, and .FALSE. if the input matrix is singular.
   real                           :: B(2,2)   ! Inverse matrix
   real                           :: det, detinv
   real, parameter                :: EPS = epsilon(1.0)

   ! Calculate the inverse determinant of the matrix
   det = determinante_22(A)
   if (abs(det) < EPS) then
       if (present(ok_flag)) ok_flag = .false.
       B(:,:) = 0
       return
   endif
   detinv =  1 / det

   ! Calculate the inverse of the matrix
   B(1,1) = +detinv * A(2,2)
   B(2,1) = -detinv * A(2,1)
   B(1,2) = -detinv * A(1,2)
   B(2,2) = +detinv * A(1,1)

   if (present(ok_flag)) ok_flag = .true.

   end function matinv2

!------------------------------------------------------------------------------------------

   function matinv3(A,ok_flag) result(B)
   !
   !Performs a direct calculation of the inverse of a 3×3 matrix.
   real, intent(in)               :: A(3,3)   ! Matrix
   logical, intent(out), optional :: ok_flag  ! .TRUE. if the input matrix could be inverted, and .FALSE. if the input matrix is singular.
   real                           :: B(3,3)   ! Inverse matrix
   real                           :: det, detinv
   real, parameter                :: EPS = epsilon(1.0)

   ! Calculate the inverse determinant of the matrix
   det = determinante(A)
   if (abs(det) < EPS) then
       if (present(ok_flag)) ok_flag = .false.
       B(:,:) = 0
       return
   endif
   detinv =  1 / det

   ! Calculate the inverse of the matrix
   B(1,1) = +detinv * (A(2,2)*A(3,3) - A(2,3)*A(3,2))
   B(2,1) = -detinv * (A(2,1)*A(3,3) - A(2,3)*A(3,1))
   B(3,1) = +detinv * (A(2,1)*A(3,2) - A(2,2)*A(3,1))
   B(1,2) = -detinv * (A(1,2)*A(3,3) - A(1,3)*A(3,2))
   B(2,2) = +detinv * (A(1,1)*A(3,3) - A(1,3)*A(3,1))
   B(3,2) = -detinv * (A(1,1)*A(3,2) - A(1,2)*A(3,1))
   B(1,3) = +detinv * (A(1,2)*A(2,3) - A(1,3)*A(2,2))
   B(2,3) = -detinv * (A(1,1)*A(2,3) - A(1,3)*A(2,1))
   B(3,3) = +detinv * (A(1,1)*A(2,2) - A(1,2)*A(2,1))

   if (present(ok_flag)) ok_flag = .true.

   end function matinv3

!------------------------------------------------------------------------------------------

   function matinv4(A,ok_flag) result(B)
   !
   !Performs a direct calculation of the inverse of a 3×3 matrix.
   real, intent(in)               :: A(4,4)   ! Matrix
   logical, intent(out), optional :: ok_flag  ! .TRUE. if the input matrix could be inverted, and .FALSE. if the input matrix is singular.
   real                           :: B(4,4)   ! Inverse matrix
   real                           :: det, detinv
   real, parameter                :: EPS = epsilon(1.0)

   ! Calculate the inverse determinant of the matrix
   det = determinante_44(A)
   if (abs(det) < EPS) then
       if (present(ok_flag)) ok_flag = .false.
       B(:,:) = 0
       return
   endif
   detinv =  1 / det

   ! Calculate the inverse of the matrix
   B(1,1) = detinv*(A(2,2)*(A(3,3)*A(4,4)-A(3,4)*A(4,3))+A(2,3)*(A(3,4)*A(4,2)   &
                    -A(3,2)*A(4,4))+A(2,4)*(A(3,2)*A(4,3)-A(3,3)*A(4,2)))
   B(2,1) = detinv*(A(2,1)*(A(3,4)*A(4,3)-A(3,3)*A(4,4))+A(2,3)*(A(3,1)*A(4,4)   &
                    -A(3,4)*A(4,1))+A(2,4)*(A(3,3)*A(4,1)-A(3,1)*A(4,3)))
   B(3,1) = detinv*(A(2,1)*(A(3,2)*A(4,4)-A(3,4)*A(4,2))+A(2,2)*(A(3,4)*A(4,1)   &
                    -A(3,1)*A(4,4))+A(2,4)*(A(3,1)*A(4,2)-A(3,2)*A(4,1)))
   B(4,1) = detinv*(A(2,1)*(A(3,3)*A(4,2)-A(3,2)*A(4,3))+A(2,2)*(A(3,1)*A(4,3)   &
                    -A(3,3)*A(4,1))+A(2,3)*(A(3,2)*A(4,1)-A(3,1)*A(4,2)))
   B(1,2) = detinv*(A(1,2)*(A(3,4)*A(4,3)-A(3,3)*A(4,4))+A(1,3)*(A(3,2)*A(4,4)   &
                    -A(3,4)*A(4,2))+A(1,4)*(A(3,3)*A(4,2)-A(3,2)*A(4,3)))
   B(2,2) = detinv*(A(1,1)*(A(3,3)*A(4,4)-A(3,4)*A(4,3))+A(1,3)*(A(3,4)*A(4,1)   &
                    -A(3,1)*A(4,4))+A(1,4)*(A(3,1)*A(4,3)-A(3,3)*A(4,1)))
   B(3,2) = detinv*(A(1,1)*(A(3,4)*A(4,2)-A(3,2)*A(4,4))+A(1,2)*(A(3,1)*A(4,4)   &
                    -A(3,4)*A(4,1))+A(1,4)*(A(3,2)*A(4,1)-A(3,1)*A(4,2)))
   B(4,2) = detinv*(A(1,1)*(A(3,2)*A(4,3)-A(3,3)*A(4,2))+A(1,2)*(A(3,3)*A(4,1)   &
                    -A(3,1)*A(4,3))+A(1,3)*(A(3,1)*A(4,2)-A(3,2)*A(4,1)))
   B(1,3) = detinv*(A(1,2)*(A(2,3)*A(4,4)-A(2,4)*A(4,3))+A(1,3)*(A(2,4)*A(4,2)   &
                    -A(2,2)*A(4,4))+A(1,4)*(A(2,2)*A(4,3)-A(2,3)*A(4,2)))
   B(2,3) = detinv*(A(1,1)*(A(2,4)*A(4,3)-A(2,3)*A(4,4))+A(1,3)*(A(2,1)*A(4,4)   &
                    -A(2,4)*A(4,1))+A(1,4)*(A(2,3)*A(4,1)-A(2,1)*A(4,3)))
   B(3,3) = detinv*(A(1,1)*(A(2,2)*A(4,4)-A(2,4)*A(4,2))+A(1,2)*(A(2,4)*A(4,1)   &
                    -A(2,1)*A(4,4))+A(1,4)*(A(2,1)*A(4,2)-A(2,2)*A(4,1)))
   B(4,3) = detinv*(A(1,1)*(A(2,3)*A(4,2)-A(2,2)*A(4,3))+A(1,2)*(A(2,1)*A(4,3)   &
                    -A(2,3)*A(4,1))+A(1,3)*(A(2,2)*A(4,1)-A(2,1)*A(4,2)))
   B(1,4) = detinv*(A(1,2)*(A(2,4)*A(3,3)-A(2,3)*A(3,4))+A(1,3)*(A(2,2)*A(3,4)   &
                    -A(2,4)*A(3,2))+A(1,4)*(A(2,3)*A(3,2)-A(2,2)*A(3,3)))
   B(2,4) = detinv*(A(1,1)*(A(2,3)*A(3,4)-A(2,4)*A(3,3))+A(1,3)*(A(2,4)*A(3,1)   &
                    -A(2,1)*A(3,4))+A(1,4)*(A(2,1)*A(3,3)-A(2,3)*A(3,1)))
   B(3,4) = detinv*(A(1,1)*(A(2,4)*A(3,2)-A(2,2)*A(3,4))+A(1,2)*(A(2,1)*A(3,4)   &
                    -A(2,4)*A(3,1))+A(1,4)*(A(2,2)*A(3,1)-A(2,1)*A(3,2)))
   B(4,4) = detinv*(A(1,1)*(A(2,2)*A(3,3)-A(2,3)*A(3,2))+A(1,2)*(A(2,3)*A(3,1)   &
                    -A(2,1)*A(3,3))+A(1,3)*(A(2,1)*A(3,2)-A(2,2)*A(3,1)))

   if (present(ok_flag)) ok_flag = .true.

   end function matinv4

 END MODULE math_util
