MODULE mpi_prog
   USE errormod

   implicit none

   integer, parameter :: mpi_prog_master = 0   ! rank of master process 
   integer            :: mpi_prog_rank = 0     ! rank of the process
   integer            :: mpi_prog_numprocs = 1 ! number of process
   type(error_type)   :: mpi_prog_err          ! error in MPI

#ifdef MPI
private :: gather_data_sl
interface gather_data
   module procedure gather_data_sl
end interface

private :: get_data_sl
interface get_data
   module procedure get_data_sl
end interface

CONTAINS 

   subroutine mpi_prog_init()
   USE mpi
   integer :: mpi_err
!
!  Initialize the MPI system
   call MPI_Init(mpi_err)
   if (mpi_err /= 0) then
       call mpi_prog_err%set('MPI SYSTEM could not be initialized',code=mpi_err)
       return
   endif
!
!  Get the rank of this process, store in mpi_prog_rank
   call MPI_Comm_rank(MPI_COMM_WORLD, mpi_prog_rank, mpi_err)
   if (mpi_err /= 0) then
       call mpi_prog_err%set('MPI SYSTEM did not return RANK',code=mpi_err)
       return
   endif
!
!  Get the number of processes, store in mpi_prog_numprocs
   call MPI_Comm_size(MPI_COMM_WORLD, mpi_prog_numprocs, mpi_err)
   if (mpi_err /= 0) then
       call mpi_prog_err%set('MPI SYSTEM did not return SIZE',code=mpi_err)
       return
   endif

   end subroutine mpi_prog_init

!----------------------------------------------------------------------------------------------

   integer function mpi_prog_ntrials(ntot)  result(ntrials)
!
!  Compute the number of trials for each process. Distribute the remainder
!
   integer, intent(in) :: ntot ! total number of trials
!
   if (mpi_prog_numprocs < ntot) then
       ntrials = ntot / mpi_prog_numprocs
!
!      balance remainder
       if (mpi_prog_rank < mod(ntot,mpi_prog_numprocs)) then
           ntrials = ntrials + 1
       endif
   else
       ntrials = 1
   endif
!
   end function mpi_prog_ntrials

!----------------------------------------------------------------------------------------------

   subroutine mpi_prog_print(kpr)
   integer, intent(in) :: kpr
!
   write(kpr,'(a,i0)')'Rank of this process: ',mpi_prog_rank
   write(kpr,'(a,i0)')'Number of processes:  ',mpi_prog_numprocs
!
   end subroutine mpi_prog_print

!----------------------------------------------------------------------------------------------

   subroutine gather_data_sl(myval,vector)
!
!  Every processes send myval to the master
!
   USE mpi
   logical, intent(in)                :: myval
   logical, dimension(:), intent(out) :: vector
   integer                            :: mpi_err
!
   call MPI_gather(myval,1,MPI_LOGICAL,vector,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpi_err)
!
   end subroutine gather_data_sl

!----------------------------------------------------------------------------------------------

   subroutine get_data_sl(val)
!
!  Master send val to all processes
!
   USE mpi
   logical :: val
   integer :: mpi_err
!
   call MPI_bcast(val,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpi_err)
!
   end subroutine get_data_sl

#endif

END MODULE mpi_prog
