module setref

  implicit none
 
  integer, parameter, private :: NLEN = 100

  type refinement_stage_type
       character(len=NLEN)                :: spar      ! code to identify the parameters to refine
       integer, dimension(:), allocatable :: vet       ! array with element type, if null all elements are refined
       integer, dimension(:), allocatable :: ph        ! array with crystal phase to refine, if null all phases are refined
       integer                            :: ncy = -1  ! number of refinement cycles. If negative, default is used
  contains
    
       procedure, public :: set_par
       procedure, public :: set_vet
       procedure, public :: set_ph
       procedure, public :: get_string
  end type

  type refinement_strategy_type
       type(refinement_stage_type), dimension(:), allocatable :: stages
       integer, private                                       :: num = 0 ! number of stages in the refinement strategy
  contains

       procedure, private :: add_stage_single
       procedure, private :: add_stage_multi

       procedure, public  :: new => new_refinement
       procedure, public  :: get_num
       generic, public    :: add_stage => add_stage_single, add_stage_multi
       procedure, public  :: prn => print_refinement
  end type

  integer, parameter, private :: STRUCT_PARAM=1, PROFILE_PARAM=2, ALL_PARAM=3
  private :: param_type
contains


   integer function numstages(stages)
   type(refinement_stage_type), dimension(:), allocatable, intent(in) :: stages
!   
   if (allocated(stages)) then
       numstages = size(stages)
   else
       numstages = 0
   endif
!
   end function numstages

!---------------------------------------------------------------------------------------------------------------------
   
   subroutine new_stages(vetr,n)
!
!  Create new stages
!
   type(refinement_stage_type), allocatable, intent(inout) :: vetr(:)
   integer, intent(in)                                     :: n

   if (n < 0) return
   if (numstages(vetr) /= n) then
       if (allocated(vetr))deallocate(vetr)
       if (n > 0) allocate(vetr(n))
   endif

   end subroutine new_stages

!---------------------------------------------------------------------------------------------------------------------

   subroutine clear_stages(vetr)
!
!  Delete all stages
!
   type(refinement_stage_type), allocatable, intent(inout) :: vetr(:)

   if (allocated(vetr)) deallocate(vetr)

   end subroutine clear_stages

!----------------------------------------------------------------------------------------------------

   subroutine resize_stages(vetr,n,savevet)
!
!  Resize array of atoms
!
   type(refinement_stage_type), allocatable, intent(inout) :: vetr(:)
   integer, intent(in)                                     :: n
   logical, optional, intent(in)                           :: savevet
   logical                                                 :: savev
   integer                                                 :: nv
   type(refinement_stage_type), allocatable                :: vsav(:)
   integer                                                 :: nsav
!
!  se n = 0 (riallocazione a 0): dealloca ed esci
   if (n == 0) then
       if (allocated(vetr)) deallocate(vetr)
       return
   endif
!
   if (.not.allocated(vetr)) then
       allocate(vetr(n))
   else
!
       nv = size(vetr)
       if (present(savevet)) then
           savev = savevet
       else
           savev = .true.
       endif
!
       if (savev) then
!
!          nsav contiene qual e' la porzione di vetr da salvare
           select case(nv-n)
             case (1:)       ! compatta x ad n
               nsav = n
             case (:-1)      ! espandi x ad n
               nsav = nv
             case (0)
               return        ! n=nv non fare niente
           end select
           allocate(vsav(n))
           vsav(:nsav) = vetr(:nsav)
           call move_alloc(vsav,vetr)
       else
           if (nv /= n) then
               deallocate(vetr)
               allocate(vetr(n))
           endif
       endif
   endif
!
   end subroutine resize_stages

!---------------------------------------------------------------------------------------------------------------------

   subroutine new_refinement(refs)
   class(refinement_strategy_type), intent(inout) :: refs
   refs%num = 0
   call clear_stages(refs%stages)
   end subroutine new_refinement

!---------------------------------------------------------------------------------------------------------------------

   integer function get_num(refs)
   class(refinement_strategy_type), intent(in) :: refs
   get_num = refs%num
   end function get_num
   
!---------------------------------------------------------------------------------------------------------------------

   subroutine add_stage_single(refs,scode,ph,ncy)
   USE elements
   USE strutil
   class(refinement_strategy_type), intent(inout) :: refs
   character(len=*), intent(in)                   :: scode
   integer, dimension(:), intent(in), optional    :: ph
   integer, intent(in), optional                  :: ncy
   integer                                        :: i,ier,endp,posat,zval
   character(len=10), dimension(150)              :: wordv
   integer                                        :: nword,nvet
   integer, allocatable, dimension(:)             :: vet
!
   ier = 0
   posat = index(scode,'@')
   if (posat > 0) then
       endp = len_trim(scode(:posat-1))
   else
       endp = len_trim(scode)
   endif
!   do i=1,endp
!      if (scode(i:i) == ' ') cycle
!      if (.not.any(code == scode(i:i))) then
!          ier = 1
!          exit
!      endif
!   enddo
!
!   if (ier == 0) then
       refs%num = refs%num+1
       call resize_stages(refs%stages,refs%num)
       call refs%stages(refs%num)%set_par(scode(1:endp))
       if (present(ncy)) refs%stages(refs%num)%ncy = ncy
       if (present(ph)) call refs%stages(refs%num)%set_ph(ph)
       if (posat > 0) then
           call get_words(scode(posat+1:),wordv,nword)
           if (nword > 0) then
               allocate(vet(nword))
               nvet = 0
               do i=1,nword
                  zval = z_from_specie(wordv(i))
                  !write(0,*)i,'W=',wordv(i),zval,trim(scode(posat+1:))
                  if (zval > 0) then
                      nvet = nvet + 1
                      vet(nvet) = zval
                  endif
               enddo
               if (nvet > 0) call refs%stages(refs%num)%set_vet(vet(:nvet))
           endif
       endif
!   endif
!
   end subroutine add_stage_single
   
!---------------------------------------------------------------------------------------------------------------------

   subroutine add_stage_multi(refs,scode,atom,radtype,ph)
   USE atom_type_util
   USE arrayutil
   USE nr
   USE arrayutil
   USE elements
   USE strutil
   class(refinement_strategy_type), intent(inout) :: refs
   character(len=*), intent(in)                   :: scode
   integer, intent(in)                            :: radtype
   integer, dimension(:), intent(in), optional    :: ph
   type(atom_type), dimension(:), allocatable     :: atom  ! required only for 
   integer                                        :: posat,nh,i
   integer, dimension(:), allocatable             :: arr
   integer, parameter                             :: HEAVY_ATOM_MIN = K_at
   type(element_type), dimension(:), allocatable :: elem

   posat = index(scode,'@') 
   !write(0,*)'POS=',posat,len_trim(scode)
   if (posat > 0 .and. posat /= len_trim(scode)) then
       call elements_from_atom(atom,elem)
       call ordina_elements(elem,RX_SOURCE)
       allocate(arr(size(elem)))
       select case (scode(posat+1:))
          case ('HEAVY')     ! select all heavy atoms
            nh = 0
            do i=1,size(elem)
               if (elem(i)%z >= HEAVY_ATOM_MIN) then
                   nh = nh + 1              
                   arr(nh) = i
               endif
            enddo
            if (nh == 0) return
            if (present(ph)) then
                call refs%add_stage(scode(:posat)//cat_svet(elem(arr(:nh))%lab,nh,sep=' '),ph)
            else
                call refs%add_stage(scode(:posat)//cat_svet(elem(arr(:nh))%lab,nh,sep=' '))
            endif

          case ('LIGHT')     ! select all heavy atoms
            nh = 0
            do i=1,size(elem)
               if (elem(i)%z < HEAVY_ATOM_MIN .and. elem(i)%z > H_at) then ! exclude H atoms
                   nh = nh + 1              
                   arr(nh) = i
               endif
            enddo
            if (nh == 0) return
            if (present(ph)) then
                call refs%add_stage(scode(:posat)//cat_svet(elem(arr(:nh))%lab,nh,sep=' '),ph)
            else
                call refs%add_stage(scode(:posat)//cat_svet(elem(arr(:nh))%lab,nh,sep=' '))
            endif
            
          case ('SEQ')
            do i=size(elem),1,-1
               if (elem(i)%z == H_at .and. (radtype == RX_SOURCE .or. radtype == ELECTRON_SOURCE)) cycle
               if (present(ph)) then
                  call refs%add_stage(scode(:posat)//trim(elem(i)%lab),ph)
               else
                  call refs%add_stage(scode(:posat)//trim(elem(i)%lab))
               endif
            enddo

       end select
   else
       if (present(ph)) then
           call refs%add_stage(scode,ph)
       else
           call refs%add_stage(scode)
       endif
   endif
   
   end subroutine add_stage_multi

!---------------------------------------------------------------------------------------------------------------------

   subroutine print_refinement(refs,kpr)
   class(refinement_strategy_type), intent(in) :: refs
   integer, intent(in)                         :: kpr
   integer                                     :: i
   do i=1,refs%num
      write(kpr,'(a,i5,a,x,a)') 'Stage ',i,':',refs%stages(i)%get_string()
   enddo
   end subroutine print_refinement

!---------------------------------------------------------------------------------------------------------------------

   subroutine set_par(stage,spar) 
   class(refinement_stage_type), intent(inout) :: stage
   character(len=*), intent(in)                :: spar
   stage%spar = spar
   end subroutine set_par

!---------------------------------------------------------------------------------------------------------------------

   subroutine set_vet(stage,vet) 
   USE arrayutil
   class(refinement_stage_type), intent(inout) :: stage
   integer, dimension(:), intent(in)           :: vet
   call new_array(stage%vet,size(vet))
   stage%vet = vet
   end subroutine set_vet

!---------------------------------------------------------------------------------------------------------------------

   subroutine set_ph(stage,vet) 
   USE arrayutil
   class(refinement_stage_type), intent(inout) :: stage
   integer, dimension(:), intent(in)           :: vet
   call new_array(stage%ph,size(vet))
   stage%ph = vet
   end subroutine set_ph

!---------------------------------------------------------------------------------------------------------------------

   function get_string(stage)
   USE arrayutil
   USE elements
   USE strutil
   class(refinement_stage_type), intent(in) :: stage
   character(len=:), allocatable            :: get_string
   integer                                  :: i,k,nword
   character(len=10), dimension(100)        :: wordv

#if 0
   select case (stage%spar)

      case ('XYZ','XYZB','B')
        get_string = stage%spar(1:1)
        do i=2,len_trim(stage%spar)
           get_string = get_string//','//stage%spar(i:i)
        enddo
        if (size_array(stage%vet) > 0) then
            get_string = get_string//' for '//specie_from_pxen(stage%vet(1))
            do i=2,size_array(stage%vet)
               get_string = get_string//', '//specie_from_pxen(stage%vet(i))
            enddo
        endif

      case default
        get_string = s_u2b(stage%spar)
        call get_words(get_string,wordv,nword)
        get_string = trim(wordv(1))
        do i=2,nword
           get_string = get_string//','//trim(wordv(i))
        enddo

   end select
#endif
   !get_string = s_u2b(stage%spar)
   call get_words(s_u2b(stage%spar),wordv,nword)
   !get_string = trim(wordv(1))
   do i=1,nword
      !if (i == 1) 
      !    get_string = trim(wordv(i))
      !else
      !    get_string = get_string//','//trim(wordv(i))
      !endif
      select case (param_type(wordv(i)))
        case (STRUCT_PARAM)
          do k=1,len_trim(wordv(i))
             if (i == 1 .and. k == 1) then
                 get_string = wordv(i)(k:k)
             else
                 get_string = get_string//','//wordv(i)(k:k)
             endif
          enddo
          if (size_array(stage%vet) > 0) then
              get_string = get_string//' for '//specie_from_pxen(stage%vet(1))
              do k=2,size_array(stage%vet)
                 get_string = get_string//', '//specie_from_pxen(stage%vet(k))
              enddo
          endif

        case (PROFILE_PARAM)
          if (i == 1) then
              get_string = trim(wordv(i))
          else
              get_string = get_string//','//trim(wordv(i))
          endif

         case (ALL_PARAM)
          get_string = 'All parameters'
     
      end select
   enddo

   end function get_string

!---------------------------------------------------------------------------------------------------------------------
!corr
!corr   subroutine get_phases(stage,numph,vet,nvet)
!corr   class(refinement_stage_type), intent(in)        :: stage
!corr   integer, intent(in)                             :: numph
!corr   integer, dimension(:), allocatable, intent(out) :: vet
!corr   integer, intent(out)                            :: nvet
!corr!
!corr!
!corr   end subroutine get_phases
!corr
!---------------------------------------------------------------------------------------------------------------------

   integer function param_type(strp)
   character(len=*), intent(in) :: strp
   select case (strp)
      case ('XYZ','XYZB','B')
        param_type = STRUCT_PARAM
      case ('ALLP')
        param_type = ALL_PARAM
      case default
        param_type = PROFILE_PARAM
   end select
   end function param_type

!---------------------------------------------------------------------------------------------------------------------

   subroutine set_rcode_atoms(atom,cell,spg,constr,stref,code,zspec)
!    
!  Per stabilire quali atomi e quali parametri affinare
!
   USE constraints
   USE unit_cell
   USE spginfom
   USE atom_basic
   USE strutil
   type(atom_type), dimension(:), intent(inout)     :: atom
   type(cell_type), intent(in)                      :: cell
   type(spaceg_type), intent(in)                    :: spg
   type(constraint_type), dimension(:), allocatable :: constr
   character(len=*), intent(in)                     :: stref ! stringa di affinamento: 'X','Y','Z','B','O' 
   integer, intent(in)                              :: code  ! =1: affina; =0: non affina
   integer, intent(in), dimension(:), optional      :: zspec ! z delle specie a cui applicare 
   integer                                          :: i   !,j
   integer                                          :: rcod
   integer                                          :: ipa
   integer                                          :: ncr
!
   do i=1,len_trim(stref)
      select case(lower(stref(i:i)))
         case ('x')
           rcod = 1

         case ('y')
           rcod = 2

         case ('z')
           rcod = 3

         case ('b')
           rcod = 4

         case ('o')
           rcod = 5

      end select
      if (present(zspec)) then
          do ipa=1,size(atom)       ! loop on atoms
             if (any(zspec == atom(ipa)%z())) then
                 ncr = par_in_constraint(constr,ipa,rcod)
!
!                Parameter involved in constraint?
                 if (ncr > 0) then
                     call refine_constraint(constr(ncr),atom,code)
                 else
                     !atom(ipa)%rcod(rcod) = code
!
!                    Don't modify negative code: parameter fixed for special position or polar axis
                     if (atom(ipa)%rcod(rcod) >= 0) atom(ipa)%rcod(rcod) = code
                 endif
             endif
          enddo
      else
          do ipa=1,size(atom)       ! loop on atoms
             ncr = par_in_constraint(constr,ipa,rcod)
!
!            Parameter involved in constraint?
             if (ncr > 0) then
                 call refine_constraint(constr(ncr),atom,code)
             else
!
!                Don't modify negative code: parameter fixed for special position or polar axis
                 if (atom(ipa)%rcod(rcod) >= 0) atom(ipa)%rcod(rcod) = code
             endif
          enddo
      endif
   enddo
!corr!
!corr!  Se affini x o y o z controlla per posizioni speciali
!corr   if (scan(lower(stref),'xyz') > 0) then
!corr       call set_lsq_conditions(atom,spg,cell)
!corr   endif

   end subroutine set_rcode_atoms

 !-----------------------------------------------------------------------

   subroutine set_lsq_conditions(atom,spg,cell)
   USE spginfom
   USE atom_type_util
   USE unit_cell
   implicit none
   type(atom_type), dimension(:), intent(inout) :: atom
   type(spaceg_type), intent(in)                :: spg
   type(cell_type), intent(in)                  :: cell
   integer                  :: i
   integer, dimension(10) :: code
!
   do i=1,size(atom)
      where(atom(i)%rcod(1:3) < 0) atom(i)%rcod(1:3) = 0  ! reset negative code
      code(:) = lsq_conditions(atom(i),spg,cell,i)
      where(code(:3) == 0) atom(i)%rcod(1:3) = -1
   enddo
!
   if (spg%polar()) then
       call spg%code_for_polar(atom(1)%rcod,-1)
   endif
!
   end subroutine set_lsq_conditions

!--------------------------------------------------------------------------------

   subroutine make_constraints(atom,legm,constr,spg,cell,riding_model,adp_factor)
!
!  Set program constraints
!
   USE unit_cell
   USE spginfom
   USE atom_type_util
   USE connect_mod
   USE constraints
!
   implicit none
   type(atom_type), dimension(:), allocatable, intent(inout)       :: atom
   type(bond_type), dimension(:), allocatable, intent(in)          :: legm
   type(constraint_type), dimension(:), allocatable, intent(inout) :: constr
   type(spaceg_type), intent(in)                                   :: spg
   type(cell_type), intent(in)                                     :: cell
   logical, intent(in)                                             :: riding_model
   real, intent(in)                                                :: adp_factor
   integer :: i
!
!  Delete program constraints
   do i=1,3
      call delete_all_constraints(constr,code=i,typec=2)
   enddo
!
!  Set constraits for special positions
   call set_constr_on_position(constr,atom,spg,cell)
   !call print_lsqcond()
   call set_lsq_conditions(atom,spg,cell)  ! and polar axis
!
!  Constraints on sof, set_lsq_conditions must be called before this
!corr   call set_constr_on_sof(atom,constr,cell)
!
   if (riding_model) then
       call add_constraints_for_riding(constr,atom,legm,spg,cell,adp_factor)
   endif
!
!  Isolate independent equations, e.g. B constr + constr of riding; B constr + constr. on site
   call check_constraints(constr)
      !call print_constraints(constr,atom,0,is_all=.true.)
!
   end subroutine make_constraints

end module setref
