MODULE timer

   USE type_constants, only: DP,I4B

   implicit none

   type timer_type 
     private
     real(DP), private                      :: sec_cpu = 0.0_dp
     real(DP), private                      :: sec_elapsed = 0.0_dp
     real(DP), private                      :: tot_cpu = 0.0_dp
     real(DP), private                      :: tot_elapsed = 0.0_dp
     integer(I4B), private                  :: nlapse = 0
     logical, private                       :: elapsed = .false.
     logical, private                       :: print_ave = .false.
     character(len=:), allocatable, private :: tname
   contains
     procedure :: reset
     procedure :: set_type
     procedure :: set_name
     procedure :: set_as_elapsed
     procedure :: set_print_ave

     procedure :: get_interval
     procedure :: get_time
     procedure :: get_average_cpu
     procedure :: get_average_elapsed

     procedure :: tic 
     procedure :: toc
     procedure :: tprint

     procedure :: print_el
     procedure :: print_cpu
   end type

   private secs_to_stringd, secs_to_stringr 
   interface secs_to_string
      module procedure secs_to_stringd, secs_to_stringr 
   end interface

!  Usage:
!
!  type(timer_type) :: t
!  
!  call t%set_type('Big procedure')
!  call t%tic()
!  ...... Run big procedure .....
!  call t%toc()
!  call t%tprint(0)
!

CONTAINS

   subroutine reset(this,tname)
   class(timer_type), intent(inout)       :: this
   character(len=*), intent(in), optional :: tname
   this%sec_cpu = 0
   this%sec_elapsed = 0
   this%tot_cpu = 0
   this%tot_elapsed = 0
   this%nlapse = 0
   if (present(tname)) call this%set_name(tname)
   end subroutine reset

!-------------------------------------------------------------------------

   subroutine set_name(this,tname)
   class(timer_type), intent(inout) :: this
   character(len=*), intent(in)     :: tname
   this%tname = tname
   end subroutine set_name

!-------------------------------------------------------------------------

   subroutine set_as_elapsed(this,el)
   class(timer_type), intent(inout) :: this
   logical, intent(in), optional    :: el
   if (present(el)) then
       this%elapsed = el
   else
       this%elapsed = .true.
   endif
   end subroutine set_as_elapsed

!-------------------------------------------------------------------------

   subroutine set_print_ave(this)
   class(timer_type), intent(inout) :: this
   this%print_ave = .true.
   end subroutine set_print_ave

!-------------------------------------------------------------------------

   real(DP) function get_average_cpu(this)  result(ave)
!
!  Set time as average
!
   class(timer_type), intent(in) :: this
!
   if (this%nlapse > 0) then
       ave = this%tot_cpu / this%nlapse
       return
   endif
   ave = 0
!
   end function get_average_cpu

!-------------------------------------------------------------------------

   real(DP) function get_average_elapsed(this)  result(ave)
!
!  Set time as average 
!
   class(timer_type), intent(in) :: this
!
   if (this%nlapse > 0) then
       ave = this%tot_elapsed / this%nlapse
       return
   endif
   ave = 0
!
   end function get_average_elapsed

!-------------------------------------------------------------------------

   subroutine set_type(this,tname,elapsed,init)
   class(timer_type), intent(inout)       :: this
   character(len=*), intent(in), optional :: tname
   logical, intent(in), optional          :: elapsed
   logical, intent(in), optional          :: init
   if (present(tname)) call this%set_name(tname)
   if (present(elapsed)) then
       call this%set_as_elapsed(elapsed)
   else
       call this%set_as_elapsed(.true.)
   endif
   if (present(init)) then
       if (init) call this%reset()
   else
       call this%reset()
   endif
   end subroutine set_type

!-------------------------------------------------------------------------

   real(DP) function  get_interval(this)
!
!  Get the last measured interval of time, call this after toc
!
   class(timer_type), intent(in) :: this
!
   get_interval = this%sec_cpu
!
   end function get_interval

!-------------------------------------------------------------------------

   real(DP) function  get_time(this)
!
!  Get time from last tic
!
   class(timer_type), intent(in) :: this
   real(DP)                  :: now
!
   call cpu_time(now)
   get_time = this%tot_cpu + (now - this%sec_cpu)
!
   end function get_time

!-------------------------------------------------------------------------

   function wtime ( )
!  Returns a reading of the wall clock time.
!  To get the elapsed wall clock time, call WTIME before and after a given
!  operation, and subtract the first reading from the second.
   integer ( kind = 4 ) clock_max
   integer ( kind = 4 ) clock_rate
   integer ( kind = 4 ) clock_reading
   real ( kind = 8 ) wtime

   call system_clock ( clock_reading, clock_rate, clock_max )

   wtime = real ( clock_reading, kind = 8 ) &
         / real ( clock_rate, kind = 8 )

   end function wtime

!-------------------------------------------------------------------------

   subroutine tic(this)
   class(timer_type), intent(inout) :: this
   call cpu_time(this%sec_cpu)   
   if (this%elapsed) this%sec_elapsed = wtime()
   end subroutine tic

!-------------------------------------------------------------------------
   
   subroutine toc(this)
   class(timer_type), intent(inout) :: this
   real(DP)                     :: now
!
!  Update CPU time
   this%nlapse = this%nlapse + 1
   call cpu_time(now)
   this%sec_cpu = now - this%sec_cpu                   ! save time of last interval
   this%tot_cpu = this%tot_cpu + this%sec_cpu
!
!  Update elapsed time
   if (this%elapsed) then
       this%sec_elapsed = wtime() - this%sec_elapsed   ! save time of last interval
       this%tot_elapsed = this%tot_elapsed + this%sec_elapsed
   endif
!
   end subroutine toc

!-------------------------------------------------------------------------

   function secs_to_stringd(time,sround) result(str)
!
!  Convert seconds in string. seconds rounded if sround is true
!
   USE strutil
   real(DP), intent(in)          :: time
   logical, intent(in),optional  :: sround
   integer                       :: hh,mm
   real                          :: secs
   character(len=:), allocatable :: str
!
   hh = int(time/3600.0)
   secs = time - 3600.0*hh
   mm = int(secs / 60.0)
   if (present(sround)) then
       if (sround) then
           str = trim(i_to_s(hh))//'h '//trim(i_to_s(mm))//'m '//trim(i_to_s(nint(secs - mm*60.0)))//'s '
       else
           str = trim(i_to_s(hh))//'h '//trim(i_to_s(mm))//'m '//trim(r_to_s(secs - mm*60.0,3))//'s '
       endif
   else
       str = trim(i_to_s(hh))//'h '//trim(i_to_s(mm))//'m '//trim(r_to_s(secs - mm*60.0,3))//'s '
   endif
!
   end function secs_to_stringd

!-------------------------------------------------------------------------

   function secs_to_stringr(time,sround) result(str)
!
!  Convert seconds in string. seconds rounded if sround is true
!
   USE strutil
   real, intent(in)              :: time
   logical, intent(in), optional :: sround
   integer                       :: hh,mm
   real                          :: secs
   character(len=:), allocatable :: str
!
   hh = int(time/3600.0)
   secs = time - 3600.0*hh
   mm = int(secs / 60.0)
   if (present(sround)) then
       if (sround) then
           str = trim(i_to_s(hh))//'h '//trim(i_to_s(mm))//'m '//trim(i_to_s(nint(secs - mm*60.0)))//'s '
       else
           str = trim(i_to_s(hh))//'h '//trim(i_to_s(mm))//'m '//trim(r_to_s(secs - mm*60.0,3))//'s '
       endif
   else
       str = trim(i_to_s(hh))//'h '//trim(i_to_s(mm))//'m '//trim(r_to_s(secs - mm*60.0,3))//'s '
   endif
!
   end function secs_to_stringr

!-------------------------------------------------------------------------

   subroutine tprint(this,kpr)
   class(timer_type), intent(inout) :: this
   integer, intent(in)              :: kpr
   call this%print_cpu(kpr,0)
   if (this%elapsed) call this%print_el(kpr,0)
   end subroutine tprint

!-------------------------------------------------------------------------

   subroutine print_el(this,kpr,maxs)
   class(timer_type), intent(in) :: this
   integer, intent(in)           :: kpr
   integer, intent(in)           :: maxs
   character(len=:), allocatable :: string
   real(DP)                      :: time
   if (this%print_ave) then
       time = this%get_average_elapsed()
   else
       time = this%tot_elapsed
   endif
   if (len_trim(this%tname) > 0) then
       if (maxs > 0) then
           allocate(character(len=maxs)::string)
           string(1:maxs) = trim(this%tname)   ! use string to have left-justified write
           write(kpr,'(1x,a)')string//' (elapsed time): '//secs_to_string(time)
       else
           write(kpr,'(1x,a)')trim(this%tname)//' (elapsed time): '//secs_to_string(time)
       endif
   else
       write(kpr,'(1x,a)')'Elapsed time: '//secs_to_string(time)
   endif
   end subroutine print_el

!-------------------------------------------------------------------------

   subroutine print_cpu(this,kpr,maxs)
   USE strutil
   class(timer_type), intent(in) :: this
   integer, intent(in)           :: kpr
   integer,           intent(in) :: maxs
   character(len=:), allocatable :: string
   real(DP)                      :: time
   if (this%print_ave) then
       time = this%get_average_cpu()
   else
       time = this%tot_cpu
   endif
   if (allocated(this%tname)) then
       if (maxs > 0) then
           allocate(character(len=maxs)::string)
           string(1:maxs) = trim(this%tname)   ! use string to have left-justified write
           write(kpr,'(1x,a)')string//' (cpu time):     '//secs_to_string(time)
       else
           write(kpr,'(1x,a)')trim(this%tname)//' (cpu time):     '//secs_to_string(time)
       endif
   else
       write(kpr,'(1x,a)')'CPU time:     '//secs_to_string(time)
   endif
   end subroutine print_cpu

!-------------------------------------------------------------------------

   subroutine print_time(timer,kpr)
!
!  Print table of times
!
   type(timer_type), dimension(:), intent(in) :: timer
   integer, intent(in)                        :: kpr
   integer                                    :: maxs,i,lent
!
!  Calculate max len of strings
   maxs = 0
   do i=1,size(timer)
      lent = len(timer(i)%tname)
      if (lent > maxs) maxs = lent
   enddo
!
   write(kpr,'(/1x,a)')'Summary of times'
   do i=1,size(timer)
      call timer(i)%print_cpu(kpr,maxs)
      call timer(i)%print_el(kpr,maxs)
   enddo
!
   end subroutine print_time

!-------------------------------------------------------------------------

   logical function diff_time(this,time)
   type(timer_type), intent(inout) :: this
   real, intent(in) :: time
   type(timer_type) :: t_save
!
   call t_save%set_as_elapsed()
   call t_save%tic()
   diff_time = t_save%sec_elapsed - this%sec_elapsed > time
       !write(0,*)'TIME=',t_save%sec_elapsed - this%sec_elapsed
   if (diff_time) this = t_save
!
   end function diff_time

END MODULE timer
